"""
operations for Git author attributions
"""

from __future__ import annotations
from pathlib import Path
import typing as T
import collections
import logging
import subprocess

from .git import gitdirs, GITEXE, TIMEOUT


def gitemail(path: Path, exclude: str = None) -> T.Iterator[tuple[Path, list[tuple[str, int]]]]:
    """
    returns email addresses of everyone who ever made a Git commit in this repo.

    Parameters
    ----------

    path : pathlib.Path
        path to Git repo
    exclude : list of str or tuple of str
        email addresses to exclude

    Yields
    ------
    d : pathlib.Path
        path to the Git repo
    emails : tuple of str, int
        email addresses with how many times they committed
    """

    for d in gitdirs(path):

        try:
            ret = subprocess.check_output(
                [GITEXE, "-C", str(d), "log", '--pretty="%ce"'], text=True, timeout=TIMEOUT
            )
        except subprocess.CalledProcessError as e:
            logging.error(f"{path}  {e}")
            continue

        addrs: T.Iterable[str] = filter(None, ret.replace('"', "").split("\n"))
        # remove blanks
        if exclude:
            addrs = (n for n in addrs if not n.startswith(exclude))

        emails = collections.Counter(addrs).most_common()

        yield d, emails



from pathlib import Path
import typing as T
import shutil


def find_matching_file(path: Path, fn: str) -> T.Iterator[Path]:
    """
    if full path file is found, return that filename

    Parameters
    ----------
    path : pathlib.Path
        top-level directory to check directories under
    fn : str
        filename to look for

    Yields
    -------
    matched : pathlib.Path
        full path to matching file
    """

    path = Path(path).expanduser().resolve()
    if not path.is_dir():
        raise NotADirectoryError(path)

    dlist = (x for x in path.iterdir() if x.is_dir())

    for d in dlist:
        if (d / fn).is_file():
            yield d


def find_dir_missing_file(path: Path, fn: str, copyfile: Path = None) -> T.Iterator[Path]:
    """
    if directory is missing a file, copy the file to that directory

    Parameters
    ----------
    path : pathlib.Path
        top-level directory to check directories under
    fn : str
        filename to look for
    copyfile : pathlib.Path, optional
        if present, copy this file into the directory that doesn't have it

    Yields
    -------
    missing : pathlib.Path
        directories missing the file
    """

    path = Path(path).expanduser().resolve()
    if not path.is_dir():
        raise NotADirectoryError(path)
    if copyfile and not isinstance(copyfile, Path):
        raise TypeError("copyfile must be Path or None")

    dlist = (x for x in path.iterdir() if x.is_dir())

    for d in dlist:
        if not (d / fn).is_file():
            if isinstance(copyfile, Path):
                shutil.copy2(copyfile, d)
                print(f"copied {copyfile} to {d}")

            yield d



"""
detect Git local repo modifications. Crazy fast by not invoking remote.

replaced by git status --porcelain:
  git ls-files -o -d --exclude-standard: # check for uncommitted files
  git --no-pager diff HEAD , # check for uncommitted work

DOES NOT WORK git log --branches --not --remotes     # check for uncommitted branches
"""

from __future__ import annotations
import argparse
import asyncio
import subprocess
import logging
from pathlib import Path

from . import _log
from .git import gitdirs, GITEXE, MAGENTA, BLACK, TIMEOUT

C0 = ["rev-parse", "--abbrev-ref", "HEAD"]  # get branch name
C1 = ["status", "--porcelain"]  # uncommitted or changed files

__all__ = ["git_porcelain"]


def git_porcelain(path: Path) -> bool:
    """
    detects if single Git repo is porcelain i.e. clean.
    May not have been pushed or fetched.

    Parameters
    ----------

    path: pathlib.Path
        path to Git repo

    Returns
    -------

    is_porcelain: bool
        true if local Git is clean
    """

    if not path.is_dir():
        raise NotADirectoryError(path)

    ret = subprocess.run(
        [GITEXE, "-C", str(path)] + C1, stdout=subprocess.PIPE, text=True, timeout=TIMEOUT
    )
    if ret.returncode != 0:
        logging.error(f"{path.name} return code {ret.returncode}  {C1}")
        return False
    return not ret.stdout


async def _git_status(path: Path) -> tuple[str, str]:
    """
    Notes which Git repos have local changes that haven't been pushed to remote

    Parameters
    ----------
    path : pathlib.Path
        Git repo directory

    Returns
    -------
    changes : tuple of pathlib.Path, str
        Git repo local changes
    """

    proc = await asyncio.create_subprocess_exec(
        *[GITEXE, "-C", str(path)] + C1, stdout=asyncio.subprocess.PIPE
    )
    stdout, _ = await proc.communicate()
    if proc.returncode != 0:
        logging.error(f"{path.name} return code {proc.returncode}  {C1}")
        return None
    out = stdout.decode("utf8", errors="ignore").rstrip()
    logging.info(path.name)
    # %% detect uncommitted changes
    if out:
        return path.name, out
    # %% detect committed, but not pushed
    proc = await asyncio.create_subprocess_exec(
        *[GITEXE, "-C", str(path)] + C0, stdout=asyncio.subprocess.PIPE
    )
    stdout, _ = await proc.communicate()
    if proc.returncode != 0:
        logging.error(f"{path.name} return code {proc.returncode}  {C0}")
        return None
    branch = stdout.decode("utf8", errors="ignore").rstrip()

    C2 = [GITEXE, "-C", str(path), "diff", "--stat", f"origin/{branch}.."]
    proc = await asyncio.create_subprocess_exec(*C2, stdout=asyncio.subprocess.PIPE)
    stdout, _ = await proc.communicate()
    if proc.returncode != 0:
        logging.error(f"{path.name} return code {proc.returncode}  {branch}")
        return None
    out = stdout.decode("utf8", errors="ignore").rstrip()
    if out:
        return path.name, out
    return None


async def git_status(path: Path, verbose: bool = False) -> list[str]:

    c = MAGENTA if verbose else ""

    changed = []
    for r in asyncio.as_completed([_git_status(d) for d in gitdirs(path)]):
        changes = await r
        if changes:
            changed.append(changes[0])
            print(c + changes[0])
            if verbose:
                print(BLACK + changes[1])

    return changed


def cli():
    p = argparse.ArgumentParser(description="get status of many Git repos")
    p.add_argument("path", help="path to look under", nargs="?", default="~/code")
    p.add_argument("-v", "--verbose", action="store_true")
    P = p.parse_args()

    _log(P.verbose)

    asyncio.run(git_status(P.path))


if __name__ == "__main__":
    cli()



import subprocess
import pytest

import gitutils.find as pgf
import gitutils.git as pg


def test_findfile(tmp_path):
    sub1 = tmp_path / "sub1"
    sub1.mkdir()
    missdir = list(pgf.find_dir_missing_file(tmp_path, "blahblah"))
    assert len(missdir) == 1
    assert missdir[0].samefile(sub1)


def test_findfile_baddir():
    with pytest.raises(NotADirectoryError):
        next(pgf.find_dir_missing_file("asdfasfdfoo", "blahblah"))


def test_missing(tmp_path):
    sub1 = tmp_path / "sub1"
    sub1.mkdir()
    miss = list(pgf.find_dir_missing_file(tmp_path, "blahblah"))
    assert len(miss) == 1
    assert miss[0].samefile(sub1)

    (sub1 / "foo").touch()
    assert len(list(pgf.find_dir_missing_file(tmp_path, "foo"))) == 0


def test_findfile_script(tmp_path):
    ret = subprocess.check_output(["find_missing_file", str(tmp_path), "blahblah"], text=True)
    assert not ret


def test_actonchanged(git_init):
    p = git_init
    ch = pg.listchanged(p)
    assert not ch
    ret = subprocess.check_output(["ActOnChanged", str(p)], text=True)
    assert not ret

    fn = p / "foo.txt"
    fn.touch()
    subprocess.check_call(["git", "-C", str(p), "add", fn])
    fn.write_text("hello")
    ch = pg.listchanged(p)
    assert ch
    ret = subprocess.check_output(["ActOnChanged", str(p)], text=True)
    assert ret



import pytest
import subprocess
import asyncio

from gitutils.branch import git_branch


@pytest.mark.parametrize("name, N", [("main", False), ("fake", True)])
def test_script_branch(name, N, git_commit):
    p = git_commit

    ret = subprocess.check_output(["gitbranch", str(p), "-main", name], text=True)

    if N:
        assert ret
    else:
        assert not ret


@pytest.mark.parametrize("name,N", [("main", 0), ("fake", 1)])
def test_branch(name, N, git_commit):
    p = git_commit
    branches = asyncio.run(git_branch(name, p))
    assert len(branches) == N, f"{branches}"



"""
Operations for Git branches

git branch get name methods:
https://stackoverflow.com/a/45028375
"""

from __future__ import annotations
import argparse
from pathlib import Path
import asyncio
import logging
import subprocess

from . import _log
from .git import GITEXE, GIT_VERSION, gitdirs

BRANCH_REV = ["rev-parse", "--abbrev-ref", "HEAD"]
BRANCH_SYM = ["symbolic-ref", "--short", "HEAD"]

# BRANCH_NAME = ["name-rev", "--name-only", "HEAD"]  # don't use--shows tag when HEAD coincides instead of branch
BRANCH_NAME = ["branch", "--show-current"]  # Git >= 2.22

SWITCH = ["switch"]  # Git >= 2.23


async def different_branch(main: list[str], path: Path) -> tuple[str, str]:
    """
    does branch not match specified name

    Parameters
    ----------

    main : str
        desired default branch name
    path : pathlib.Path
        Git repo to check

    Returns
    -------

    branch : tuple of pathlib.Path, str
        repo path and branch name
    """

    minver = 2.23

    if GIT_VERSION < minver:
        raise EnvironmentError(
            f"Git >= {minver} needed to use branch operations. Your Git: {GIT_VERSION}"
        )

    proc = await asyncio.create_subprocess_exec(
        *[GITEXE, "-C", str(path)] + BRANCH_NAME, stdout=asyncio.subprocess.PIPE
    )
    stdout, _ = await proc.communicate()
    if proc.returncode != 0:
        logging.error(f"{path.name} return code {proc.returncode}  {BRANCH_NAME}")
    logging.info(str(path))

    branch_name = stdout.decode("utf8").strip()

    if isinstance(main, str):
        main = [main]

    if branch_name not in main:
        return path.name, branch_name

    return None


async def git_branch(branch: list[str], path: Path) -> list[tuple[str, str]]:

    different = []
    for r in asyncio.as_completed([different_branch(branch, d) for d in gitdirs(path)]):
        diff = await r
        if diff:
            different.append(diff)
            print(diff)

    return different


def branch_switch(path: Path, old_branch: str, new_branch: str):

    for d in gitdirs(path):
        cmd = [GITEXE, "-C", str(d)] + BRANCH_NAME
        current = subprocess.check_output(cmd, text=True, timeout=5).strip()
        if current == new_branch:
            continue
        if current != old_branch:
            logging.info(f"not changing: {d.name}: {current} != {old_branch}")
            continue

        cmd = [GITEXE, "-C", str(d)] + SWITCH + [new_branch]
        ret = subprocess.run(
            cmd,
            stderr=subprocess.PIPE,
            timeout=10,
            text=True,
        )
        if ret.returncode != 0:
            stderr = ret.stderr.strip()
            if stderr in {
                f"fatal: invalid reference: {new_branch}",  # switch
                f"error: pathspec '{new_branch}' did not match any file(s) known to git",  # checkout
            }:
                continue
            else:
                raise ValueError(
                    f"{d} could not switch to {new_branch} from {old_branch}: {stderr}"
                )

        print(d.name, old_branch, "=>", new_branch)


def cli():
    """
    report on git repos not on the expected branch e.g. 'master'
    """

    p = argparse.ArgumentParser(
        description="check for non-default branches, and mass switch branches"
    )
    p.add_argument("path", help="path to look under", nargs="?", default="~/code")
    p.add_argument("-v", "--verbose", action="store_true")
    p.add_argument("-main", nargs="+", default=["main", "master"], help="name of your main branch")
    p.add_argument("-switch", nargs=2, help="switch branch from OLD to NEW, if NEW is available")
    P = p.parse_args()

    _log(P.verbose)

    if P.switch is None:
        asyncio.run(git_branch(P.main, P.path))
    else:
        branch_switch(P.path, *P.switch)


if __name__ == "__main__":
    cli()



"""
These Git utilities use nothing beyond pure Python and command-line Git.
Speed is emphasized throughout, with pipelining and concurrent `asyncio` routines throughout
for fastest operation on large numbers of repos.
"""

import logging


def _log(verbose: bool):

    if verbose:
        logging.basicConfig(level=logging.DEBUG)



import argparse
from pathlib import Path
import webbrowser

from . import _log
from .git import MAGENTA, BLACK, listchanged
from .find import find_matching_file, find_dir_missing_file
from .email import gitemail

p = argparse.ArgumentParser()
p.add_argument("path", help="path to look under", nargs="?", default="~/code")
p.add_argument("-v", "--verbose", action="store_true")


def find_match():
    p.add_argument("fn", help="filename to look for")
    P = p.parse_args()

    _log(P.verbose)

    for f in find_matching_file(P.fn, P.path):
        print(f)


def ActOnChanged():
    # needs to have its own argparser
    p = argparse.ArgumentParser()
    p.add_argument("path", help="root path to search under", nargs="?", default=".")
    p.add_argument("-p", "--preview", help="web browser preview of localhost", action="store_true")
    p.add_argument(
        "--port",
        help="port of localhost web server (Jekyll: 4000, Hugo: 1313)",
        type=int,
        default=1313,
    )
    p.add_argument("-v", "--verbose", action="store_true")
    P = p.parse_args()

    _log(P.verbose)

    path = Path(P.path).expanduser().resolve()
    flist = listchanged(path)
    # %%
    if P.preview:
        prefix = f"http://localhost:{P.port}/"

        if path.name == "_posts":  # Jekyll with leading date in filename
            cut = 11
        else:
            cut = 0
        flist = (prefix + fn.split("/")[-1][cut:].split(".")[0] for fn in flist)

        for f in flist:
            webbrowser.open_new_tab(f)
    else:
        for f in flist:
            print(path / f)


def find_missing():
    p.add_argument("fn", help="filename to look for")
    p.add_argument("-c", "--copy", help="filepath to copy in if missing")
    P = p.parse_args()

    _log(P.verbose)

    filedest = Path(P.copy).expanduser() if P.copy else None

    for f in find_dir_missing_file(P.path, P.fn, filedest):
        print(f)


def git_email():
    p.add_argument("-e", "--exclude", help="user to ignore (keep)")
    P = p.parse_args()

    _log(P.verbose)

    for d, emails in gitemail(P.path, P.exclude):
        print(MAGENTA + d.stem + BLACK)
        for email in emails:
            print(*email)



import subprocess
import pytest


@pytest.mark.parametrize("op", ["gitemail"])
def test_email(op, git_commit):
    p = git_commit
    ret = subprocess.check_output([op, str(p)], text=True)
    assert ret



"""
Git fetch / pull functions
"""

from __future__ import annotations
import argparse
import asyncio
import subprocess
import logging
from pathlib import Path
import urllib.request
import socket

from . import _log
from .git import GITEXE, gitdirs


def check_internet() -> bool:
    """
    check that Internet connection is working
    """

    url = "https://connectivitycheck.gstatic.com/generate_204"

    try:
        with urllib.request.urlopen(url, timeout=3) as resp:
            if resp.getcode() == 204:
                return True
    except (socket.gaierror, urllib.error.URLError):
        raise ConnectionError(f"No internet connection to {url}")

    return False


async def fetchpull(mode: str, path: Path) -> Path:
    """
    handles recursive "git pull" and "git fetch"

    Parameters
    ----------

    mode : str
        fetch or pull
    path : pathlib.Path
        Git repo path

    Returns
    -------
    failed : pathlib.Path
        Git repo that failed to fetch/pull


    Reference:
    ----------
    format mini-language:
    https://docs.python.org/3/library/string.html#format-specification-mini-language


    Note: Don't use git pull --quiet because you get no output at all when remote change
    occured. Leave it as is with stdout=DEVNULL and no --quiet.
    """

    cmd = [GITEXE, "-C", str(path), mode]

    proc = await asyncio.create_subprocess_exec(
        *cmd,
        stdout=subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
        stdin=asyncio.subprocess.DEVNULL,
    )
    stdout, stderr = await proc.communicate()
    out = stdout.decode("utf8", errors="ignore").rstrip()
    if out and out != "Already up to date.":
        print(path.name, out)

    if mode == "fetch" and proc.returncode == 0:
        cmd = [GITEXE, "-C", str(path), "diff", "--stat", "HEAD..FETCH_HEAD"]
        proc = await asyncio.create_subprocess_exec(
            *cmd,
            stdout=subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            stdin=asyncio.subprocess.DEVNULL,
        )
        stdout, stderr = await proc.communicate()
        out = stdout.decode("utf8", errors="ignore").rstrip()
        if out:
            print(path.name, out)
    err = stderr.decode("utf8", errors="ignore").rstrip()
    if proc.returncode:
        if "Permission denied" in err or "fatal: could not read Password" in err:
            logging.info(f"SKIP: credentials needed: {path.name}")
            return None

        logging.error(f"{path.name}  {err}")
        return path

    logging.info(f"{mode} {path.name}")
    return None


async def git_pullfetch(mode: str, path: Path) -> list[Path]:

    failed = []
    for r in asyncio.as_completed([fetchpull(mode, d) for d in gitdirs(path)]):
        fail = await r
        if fail:
            failed.append(fail)
            print(fail.name)

    return failed


def git_fetch_cli():
    p = argparse.ArgumentParser()
    p.add_argument("path", help="path to look under", nargs="?", default="~/code")
    p.add_argument("-v", "--verbose", action="store_true")
    P = p.parse_args()

    _log(P.verbose)

    asyncio.run(git_pullfetch("fetch", P.path))


def git_pull_cli():
    p = argparse.ArgumentParser()
    p.add_argument("path", help="path to look under", nargs="?", default="~/code")
    p.add_argument("-v", "--verbose", action="store_true")
    P = p.parse_args()

    _log(P.verbose)

    if not check_internet():
        raise ConnectionError("No internet connection")

    asyncio.run(git_pullfetch("pull", P.path))


if __name__ == "__main__":
    git_pull_cli()



import pytest
import subprocess


@pytest.fixture(scope="function")
def git_init(tmp_path):
    subprocess.check_call(["git", "-C", str(tmp_path), "init", "-b", "main"])
    return tmp_path


@pytest.fixture(scope="function")
def git_commit(git_init):
    p = git_init
    (p / "foo.txt").touch()
    subprocess.check_call(["git", "-C", str(p), "add", "."])

    subprocess.check_call(["git", "-C", str(p), "config", "user.email", "you@example.invalid"])
    subprocess.check_call(["git", "-C", str(p), "config", "user.name", "Your Name"])

    subprocess.check_call(["git", "-C", str(p), "commit", "-am", "test", "--no-verify"])
    return p



from pathlib import Path
import subprocess
import pytest
import asyncio

from gitutils.pull import git_pullfetch


@pytest.mark.skipif(__file__ is None, reason="can't find own directory")
@pytest.mark.parametrize("op", ["gitpull", "gitfetch"])
def test_script_pull(op):
    R = Path(__file__).resolve().parent
    ret = subprocess.check_output([op, str(R)], text=True).strip()
    assert not ret


@pytest.mark.parametrize("mode", ["fetch", "pull"])
def test_nonGit_dir_pull(tmp_path, mode):

    dirs = asyncio.run(git_pullfetch(mode, tmp_path))
    assert len(dirs) == 0
